﻿/* Program.cs
   Part of "APCComm / windows Version" 
   Copyright (C) 2000,2001,2006 Ralf Hoffmann
   Original Contact: ralf@boomerangsworld.de
   Windows version Contact: evil.lombo@gmail.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/
/* $Id: Program.cs,v 1.11 2013/10/30 22:35:32 Krahen Exp $ */

using System;
using System.Diagnostics;
using System.IO;
using System.Text;
using System.Security.Principal;


namespace APCComm
{
	
	enum apc_command_t { APC_COMMAND_FILETRANSFER=0,
		APC_COMMAND_ERROR,
		APC_COMMAND_ENTERDIR,
		APC_COMMAND_LEAVEDIR,
		APC_COMMAND_FINISHED,
		APC_COMMAND_NOP,
		APC_COMMAND_ABORT,
		APC_COMMAND_SKIPFILE } ;
	
	struct transfer_log {
		public long files;
		public ulong bytes;
		public TimeSpan tvs;
	};
	
	struct pc_file_tag {
		public FileStream fp;
		public int size;
		public uint checksum;
		public String name;
	};
	
	enum overwrite_mode_t { OVERWRITE_REQUEST, OVERWRITE_ALL, OVERWRITE_NONE };
	
	class Program
	{
		const int TRANSFERBLOCKSIZE = 1024;
		const int MAJOR = 2;
		const int MINOR = 3;
		const int PATCH = 0;
		
		static byte[] outblock = new Byte[TRANSFERBLOCKSIZE];
		static byte[] inblock = new Byte[TRANSFERBLOCKSIZE];
		static byte data = 0;
		
		static ParallelPort.PortAccess dataPort;
		static byte status;
		static ParallelPort.PortAccess statusPort;
		static int amiga_nibble;
		static String basedir;
		static transfer_log tfl;
		static pc_file_tag infile, outfile;
		static overwrite_mode_t overwrite_mode = overwrite_mode_t.OVERWRITE_REQUEST;
		static bool x64BitSystem;
		static bool x64BitProcess;
	
		public static void Main(string[] args)
		{
			short port = 888;
			bool opt = false;
			uint o1,o2,o3;
			
			x64BitSystem = Environment.Is64BitOperatingSystem;
			x64BitProcess = Environment.Is64BitProcess;
			
			if (x64BitSystem != x64BitProcess) {
				Console.WriteLine("You are running in 32-bit mode on a 64-bit operating system. Please run the x64 version.");
				return;
			}
			
			if (!IsUserAdministrator()){
				Console.WriteLine("You need to run the process with Administrator privileges.");
				return;
			}
			
			/* Lower the priority so it does not interfere with your daily job*/
			using (Process p = Process.GetCurrentProcess()){
				p.PriorityClass = ProcessPriorityClass.Idle;
			}
			
			if (args.Length > 0) {
				if (args[0].ToLowerInvariant().StartsWith("port:")){
					opt = true;
					try{
						port = short.Parse(args[0].Substring(5));
					}
					catch (Exception){
						Console.WriteLine("The port should be base10 integer. Defaulting to 888 (378h)");
					}
				}
			}
			else{
				Console.WriteLine("No port specified. Defaulting to 888 (378h)");
			}
			dataPort = new ParallelPort.PortAccess(port,x64BitSystem);
			statusPort = new ParallelPort.PortAccess((short)(port+1),x64BitSystem);
			status = statusPort.Read();
			Console.WriteLine("Waiting for Amiga:");
			synchro();
			Console.WriteLine("\r\nConnection established\r\n");
			putint(ref outblock,MAJOR,0);
			putint(ref outblock,MINOR,4);
			putint(ref outblock,PATCH,8);
			if (args.Length > 0 )
				putint(ref outblock,1,12);
			else
				putint(ref outblock,0,12);
			/* Inform the other side */
			transferblock();
			
			/* Now check the other version */
			o1=getint(inblock,0);
			o2=getint(inblock,4);
			o3=getint(inblock,8);

			if( (o1!=MAJOR) || (o2!=MINOR) ) {
				Console.WriteLine("Wrong version at the other side!");
				Console.WriteLine("  Expected 2.3.x ,  got "+o1+"."+o2+"."+o3);
				return;
			}
			
			if( (o1==0) && (args.Length==0) ) {
				Console.WriteLine("Nothing to do!\n");
				return;
			}
			
			if ((args.Length > 1 && opt) || (!opt && args.Length>0))
				send(args);
			else
				receive();
		}
		
		static void setack(bool value){
			byte pin = (1 << 1);
			if (value){
				data |= pin;
			}
			else{
				pin = (byte)((~ pin)&0xFF);
				data &= pin;
			}
			dataPort.Write(data);
		}
		
		static void putvalue(int value)
		{
			byte pin;
			if(value==3){ pin = (1 << 4) | (1 << 2); data |= pin;}
			else if(value==0){ pin = (~((1 << 4) | (1 << 2)))&0xFF; data &= pin;}
			else {
				if ((value & 1) != 0){
					pin = (1 << 4);
					data |= pin;
				}
				else{
					pin = (~(1 << 4))&0xFF;
					data &= pin;
				}
				
				if ((value & 2) != 0){
					pin = (1 << 2);
					data |= pin;
				}
				else{
					pin = (~(1 << 2)) & 0xFF;
					data &= pin;
				}
			}
			dataPort.Write(data);
		}
		
		static bool getack()
		{
			bool ack;
			status = statusPort.Read();
			ack =  (status & ( 1 << 7 )) == 0;
			return ack;
		}
		
		static int getnibble()
		{
			bool pin10,pin12,pin13,pin15;
			int i2;
			status = statusPort.Read();
			pin10 = (status & ( 1 << 6 )) != 0;
			pin12 = (status & ( 1 << 5 )) != 0;
			pin13 = (status & ( 1 << 4 )) != 0;
			pin15 = (status & ( 1 << 3 )) != 0;
			if(pin10) i2=8; else i2=0;
			if(pin12) i2|=4;
			if(pin13) i2|=2;
			if(pin15) i2|=1;
			amiga_nibble=i2;
			return i2;
		}
		
		static void synchro()
		{
			int count;
			setack(true);
			putvalue(0);
			for(count=0;!getack();count++) {
				System.Threading.Thread.Sleep(1000);
				if(count%100==0) {
					Console.Write(".");
				}
			}
			setack(false);
			for(count=0;getnibble()!=1;count++) {
				System.Threading.Thread.Sleep(1000);
				if(count%100==0) {
					Console.Write(".");
				}
			}
			putvalue(1);
			for(count=0;getnibble()!=2;count++) {
				System.Threading.Thread.Sleep(1000);
				if(count%100==0) {
					Console.Write(".");
				}
			}
			putvalue(2);
			setack(true);
		}
		
		static void putint(ref byte[] buffer,uint value, int offset)
		{
			uint b3,b2,b1,b0;
			b0=value&0xff;
			value>>=8;
			b1=value&0xff;
			value>>=8;
			b2=value&0xff;
			value>>=8;
			b3=value&0xff;
			buffer[offset] = (byte)b3;
			buffer[offset+1] = (byte)b2;
			buffer[offset+2] = (byte)b1;
			buffer[offset+3] = (byte)b0;
		}

		static void transferblock(){
			int i,s1;
			int outbyte=0,inbyte,bytecyc,ti;

			bytecyc=3;
			for(i=0;i<TRANSFERBLOCKSIZE;i++) {
				s1=0;
				if(bytecyc==3) outbyte=outblock[i>>1];
				s1=outbyte>>((bytecyc--)*2);
				while(getack());
				putvalue(s1&0x3);
				inbyte=getnibble();
				setack(false);
				s1=outbyte>>((bytecyc--)*2);
				while(!getack());
				putvalue(s1&0x3);
				ti=getnibble();
				ti<<=4;
				inbyte|=ti;
				setack(true);
				if(bytecyc<0) bytecyc=3;
				inblock[i]=(byte)inbyte;
			}
		}
		
		static uint getint(byte[] buffer, int offset)
		{
			uint value;
			value=(uint)buffer[offset+0];
			value<<=8;
			value|=(uint)buffer[offset+1];
			value<<=8;
			value|=(uint)buffer[offset+2];
			value<<=8;
			value|=(uint)buffer[offset+3];
			return value;
		}
		
		static int init_incoming_file(String name){
			FileStream fp;
			String basename;
			String tstr;
			String choose, newname;
			int fileok;
			
			infile.size=0;
			infile.checksum=0;
			infile.fp= null;
			
			if (basedir!= null){
				basename = basedir;
			}
			else {
				basename = "";
			}
			
			newname = name;
			fileok=0;
			do {
				tstr = basename+newname;
				Console.Write(basename + newname);
				if (File.Exists(basename+newname)){
					if ( overwrite_mode == overwrite_mode_t.OVERWRITE_ALL ) {
						fileok = 1;
					} else if ( overwrite_mode == overwrite_mode_t.OVERWRITE_NONE ) {
						fileok = -1;
					} else {
						Console.Write("File "+tstr+" already exists!\n\tOverwrite (o), Overwrite all (o!),\n\tSkip (s), Skip all(s!),\n\tchange (c) name or abort (a):");
						choose = Console.ReadLine();
						if (choose[0].Equals('o')){
							fileok=1;
							if ( choose.Length>1 && choose[1] == '!' ) overwrite_mode = overwrite_mode_t.OVERWRITE_ALL;
						} else if ( choose[0] == 's' ) {
							fileok = -1;
							if ( choose.Length>1 && choose[1] == '!' ) overwrite_mode = overwrite_mode_t.OVERWRITE_NONE;
						}else if ( choose[0] == 'a' ) {
							fileok = -2;
						} else {
							Console.Write("Enter new name:");
							newname = Console.ReadLine();
						}
					}
				} else fileok=1;
				
			} while(fileok==0);
			if ( fileok < 0 ) {
				return fileok;
			}
			fp = File.OpenWrite(tstr);
			infile.fp = fp;
			if (fp == null) {
				return 1;
			}
			infile.name = newname;
			return 0;
		}
		
		static void close_incoming_file( )
		{
			if(infile.fp!=null)
				infile.fp.Close();
		}
		
		static int recvfile(){
			DateTime tvs, tvi;
			tvs=tvi=new DateTime();
			double msecs;
			int checksum,wrong,s1;
			uint inchecksum;
			uint tc, l, mode, size = 0;
			double d1;
			int getfile;
			int recbytes,count;
			String outbuf,recvbuf;
			String cwd;
			byte[] tstr=null;
			byte[] inarg=null;
			int ret;

			size=getint(inblock,4);
			l=getint(inblock,8);
			
			inarg = new byte[l];
			Buffer.BlockCopy(inblock,12,inarg,0,(int)l);
			
			mode=getint(inblock,TRANSFERBLOCKSIZE-4);
			Console.WriteLine("receiving file \""+Encoding.UTF8.GetString(inarg)+"\", size "+size);
			ret = init_incoming_file(Encoding.UTF8.GetString(inarg));
			
			getfile=1;

			checksum=0;

			outbuf = size.ToString();
			recvbuf = "RECV: {0:0}/{1:0} ({2:0.00%})";
			
			inchecksum=0;
			recbytes=0;
			wrong=0;
			tc = 0;

			if((ret>0)||(ret==-2)) {
				Console.WriteLine( "Can't create "+inarg+" in "+basedir+", aborting\n");
				sendcommand(apc_command_t.APC_COMMAND_ABORT,null, 0 );
			} else if(ret==-1) {
				sendcommand(apc_command_t.APC_COMMAND_SKIPFILE,null, 0 );
				ret = 0;
			} else if(ret==0) {
				sendcommand(apc_command_t.APC_COMMAND_NOP,null, 0 );
				tvs = DateTime.Now;
				for(count=0;getfile==1;count++) {
					if(recbytes<size) {
						if(wrong==0) putint(ref outblock,0,TRANSFERBLOCKSIZE/2-4);
						else putint(ref outblock,0xffffffff,TRANSFERBLOCKSIZE/2-4);

						if(getfile==1) {
							transferblock();
							tc=getint(inblock,TRANSFERBLOCKSIZE-4);
							if(tc==0xffffffff) wrong=2;
							else if ( tc == 0xffffffee ) wrong = 4;
						}
						if(wrong!=0) break;
						
						s1=(int)size-recbytes;
						if(s1>TRANSFERBLOCKSIZE-4) s1=TRANSFERBLOCKSIZE-4;
						else getfile=2;
						
						writefrombuffer(ref inblock,s1,ref inchecksum);
						recbytes+=s1;
						checksum=(int)(inchecksum-tc);
						if(checksum!=0) wrong=1;
					}
					else getfile = 2;
					
					if ( ( count == 19 ) || ( getfile != 1 ) ) {
						count=0;
						if ( size != 0)
							outbuf=String.Format(recvbuf,recbytes,size,(double)recbytes/size);
						else
							outbuf=String.Format(recvbuf,0,0,100.0);
						Console.Write("\r"+outbuf);
					}
				}
				tvi = DateTime.Now;
				Console.WriteLine("\r\nTransfer finished");
				if(wrong==1) {
					Console.WriteLine();
					Console.WriteLine("Error while transfer\r\nI have received wrong data!");
					Console.WriteLine("Please repeat transmission!");
				}
			} else if(wrong==2) {
				Console.WriteLine();
				Console.WriteLine("Error while transfer\r\nThe Amiga reports data loss!");
				Console.WriteLine("Please repeat transmission!");
			}else if ( wrong == 4 ) {
				ret = -1;
			} else {
				if(getfile>0) {
					if (tvi != null && tvs != null){
						msecs = (tvi - tvs).TotalMilliseconds;
					}
					else {
						msecs = 0;
					}
					d1=size;
					d1*=1000;
					d1/=msecs;
					d1/=1024;
					Console.WriteLine(String.Format(", file received with {0:0.00} KBytes/s",d1));
					
					tfl.files++;
					tfl.bytes += size;
					tfl.tvs.Add(new TimeSpan(0,0,0,0,(int)msecs));
				}
				
			}
			close_incoming_file();
			/* File is stored in the current dir when basedir is NULL
			 * File is stored relative to current dir in basedir when basedir starts NOT with "X:\\"
			 * File is stored in basedir when basedir starts with "X:\"
			 */

			/* First get the cwd */
			cwd = Directory.GetCurrentDirectory();
			if ( basedir == null ) {
				if (cwd.Length < 4) {
					tstr = new byte[cwd.Length+infile.name.Length+1];
					tstr = Encoding.UTF8.GetBytes(cwd+infile.name);
				}
				else {
					tstr = new byte[cwd.Length+infile.name.Length+1];
					tstr = Encoding.UTF8.GetBytes(cwd+"\\"+infile.name);
				}
			}
			else {
				
				/* at least relative to basedir */
				if (!basedir[1].Equals(':') ){
					/* basedir is relative */
					if (cwd.EndsWith("\\")){
						tstr =Encoding.UTF8.GetBytes(cwd+ basedir+ infile.name);
					}
					else {
						tstr =Encoding.UTF8.GetBytes(cwd+ "\\"+basedir+ infile.name);
					}
				}
				else
					tstr = Encoding.UTF8.GetBytes(basedir+ infile.name);
			}
			
			if(tstr!=null) {
				Console.WriteLine("Incoming file is stored at: "+Encoding.UTF8.GetString(tstr));
				if (wrong != 0){
					Console.WriteLine("Wrong file is at "+Encoding.UTF8.GetString(tstr));
					ret =1;
				}
			}
			return ret;
		}
		
		static int writefrombuffer(ref byte[] buffer,int size,ref uint return_checksum)
		{
			int writed,i;
			uint checksum;
			writed = 0;
			infile.fp.Write(buffer,0,size);
			checksum=0;
			for(i=0;i<size;i++) checksum+=buffer[i];
			infile.checksum+=checksum;
			return_checksum=checksum;
			return writed;
		}

		
		static int recvcommand(ref apc_command_t com, ref byte[] arg,ref uint mode)
		{
			apc_command_t tc;
			byte[] narg;

			transferblock();
			tc=(apc_command_t)Enum.Parse(typeof(apc_command_t),getint(inblock,0).ToString());
			switch(tc) {
				case apc_command_t.APC_COMMAND_ERROR:
					com=tc;
					arg=null;
					break;
				case apc_command_t.APC_COMMAND_ENTERDIR:
					narg=new byte[TRANSFERBLOCKSIZE-4+1];
					if(narg!=null) {
						com=tc;
						Buffer.BlockCopy(inblock,4,narg,0,TRANSFERBLOCKSIZE-4-4);
						narg[TRANSFERBLOCKSIZE-4-4]=(byte)'\0';
						arg=narg;
						mode=getint(inblock,TRANSFERBLOCKSIZE-4);
					} else return 1;
					break;
				case apc_command_t.APC_COMMAND_LEAVEDIR:
					com=tc;
					arg=null;
					break;
				case apc_command_t.APC_COMMAND_FILETRANSFER:
					com=tc;
					arg=null;
					return recvfile();
				case apc_command_t.APC_COMMAND_FINISHED:
					com=tc;
					arg=null;
					break;
				case apc_command_t.APC_COMMAND_NOP:
					com=tc;
					arg=null;
					break;
				case apc_command_t.APC_COMMAND_ABORT:
					com=tc;
					arg=null;
					break;
				case apc_command_t.APC_COMMAND_SKIPFILE:
					com=tc;
					arg=null;
					break;
				default:
					Console.WriteLine( "Unknown command "+tc+" received!");
					return 1;
			}
			return 0;
		}
		
		static int init_outgoing_file(String name)
		{
			FileStream fp;
			long size;
			outfile.size=0;
			outfile.checksum=0;
			outfile.fp =null;
			
			if (File.Exists(name)){
				size = new FileInfo(name).Length;
				if (size > int.MaxValue)
					Console.WriteLine("Too big file "+name+". I'll truncate it.");
				outfile.size = (int)size;
				fp = File.OpenRead(name);
			}
			else
				return 1;
			outfile.name= name;
			outfile.fp = fp;
			return 0;
			
		}

		static int readtobuffer(ref byte[] buffer, int maxsize, ref uint return_checksum){
			int readed,i;
			uint checksum;
			readed=outfile.fp.Read(buffer,0,maxsize);
			checksum=0;
			for(i=0;i<readed;i++) checksum+=(uint)buffer[i];
			outfile.checksum+=checksum;
			return_checksum=checksum;
			return readed;
		}
		
		static int sendfile(String name, uint mode){
			DateTime tvs,tvo;
			double msecs;
			int wrong;
			uint tc,outchecksum;
			double d1;
			int sendfile;
			int sendbytes,ts,count;
			String outbuf,sendbuf;
			byte[]arg=null;
			int pos=0;
			String outbasename;
			uint amode = 0;
			
			apc_command_t com;
			int ret=0;
			
			if(init_outgoing_file(name)!=0) {
				Console.WriteLine("can't open file "+name+"!");
				return 1;
			}
			else {
				sendfile=1;
				Console.WriteLine("sending file \""+name+"\", size "+outfile.size);
			}
			
			putint(ref outblock,(uint)apc_command_t.APC_COMMAND_FILETRANSFER,0);
			putint(ref outblock,(uint)outfile.size,4);
			
			if (name.IndexOf("\\") < 0 ){
				outbasename = name;
			}
			else{
				outbasename = name.Substring(name.LastIndexOf("\\")+1);
			}
			putint(ref outblock,(uint)Math.Min(outbasename.Length,TRANSFERBLOCKSIZE/2-16),8);
			Buffer.BlockCopy(Encoding.UTF8.GetBytes(outbasename+'\0'),0,outblock,12,Math.Min(TRANSFERBLOCKSIZE/2-16,outbasename.Length+1));
			putint(ref outblock,mode,TRANSFERBLOCKSIZE/2-4);
			transferblock();
			com = (apc_command_t) getint( inblock,0 );
			if ( com != apc_command_t.APC_COMMAND_ABORT ) {
				outbuf = outfile.size.ToString();
				sendbuf = "SEND: {0:0}/{1:0} ({2:0.00%})";
				pos=6+outbuf.Length*2+1+2+6+1+1;
				
				outchecksum=0;
				sendbytes=0;
				wrong=0;
				
				/* now synchronize with other side and check state */
				if(recvcommand(ref com,ref arg,ref amode)==0) {
					if(com==apc_command_t.APC_COMMAND_NOP) {
						outchecksum=0;
						sendbytes=0;
						wrong=0;
						tvs = DateTime.Now;
						for(count=0;sendfile==1;count++) {
							ts=readtobuffer(ref outblock,TRANSFERBLOCKSIZE/2-4,ref outchecksum);
							sendbytes+=ts;
							if ( ts < 1 ) sendfile=2;
							if ( ts < 0 ) wrong = 3;
							
							if(wrong==0) putint(ref outblock,outchecksum,TRANSFERBLOCKSIZE/2-4);
							else  putint(ref outblock,0xffffffff,TRANSFERBLOCKSIZE/2-4);
							if ( ( sendfile == 1 ) || ( wrong != 0 ) ) {
								transferblock();
								tc=getint(inblock,TRANSFERBLOCKSIZE-4);
								if(tc==0xffffffff) wrong=2;
								else if ( tc == 0xffffffee ) wrong = 4; /* abort */
							}
							if(wrong!=0) break;
							
							if ( ( count == 19 ) || ( sendfile != 1 ) ) {
								count=0;
								if (outfile.size != 0)
									outbuf = String.Format(sendbuf,sendbytes,outfile.size,(double)sendbytes/outfile.size);
								else
									outbuf = String.Format(sendbuf,0,0,100.0);
								Console.Write("\r"+outbuf);
								
							}
						}
						tvo = DateTime.Now;
						Console.WriteLine("\r\nTransfer finished");
						if(wrong==1) {
							Console.WriteLine();
							Console.WriteLine("Error while transfer\r\nI have received wrong data!");
							Console.WriteLine("Please repeat transmission!");
							Console.WriteLine("Wrong file was "+name);
						}
						else if(wrong==2) {
							Console.WriteLine();
							Console.WriteLine("Error while transfer\r\nThe Amiga reports data loss!");
							Console.WriteLine("Please repeat transmission!");
							Console.WriteLine("Wrong file was "+name);
						}
						else if (wrong==3) {
							Console.WriteLine();
							Console.WriteLine("Error while reading file");
							Console.WriteLine("Wrong file was "+name);
						}
						else  if ( wrong == 4 ) {
							ret = -1;
						}
						else{
							msecs = (tvo-tvs).TotalMilliseconds;
							d1=outfile.size;
							d1*=1000;
							d1/=msecs;
							d1/=1024;
							Console.WriteLine(String.Format(", file sent with {0:0.00} KBytes/s",d1));
							tfl.files++;
							tfl.bytes += (ulong)outfile.size;
							tfl.tvs.Add(new TimeSpan(0,0,0,0,(int)msecs));
						}
					} else if(com==apc_command_t.APC_COMMAND_SKIPFILE) {
						Console.WriteLine("Skipping file!");
					}
					else if (com==apc_command_t.APC_COMMAND_ABORT) {
						ret = -1;
					}
					else {
						Console.WriteLine("Wrong command "+com);
						ret = 1;
					}
				} else {
					Console.WriteLine("Error while receiving command!");
					ret=1;
				}
			} else {
				ret = -1;
			}
			close_outgoing_file();
			return ret;
		}
		
		static void close_outgoing_file(){
			if (outfile.fp != null) outfile.fp.Close();
			
		}
		
		static int sendcommand(apc_command_t com,byte[] arg, uint mode)
		{
			if(com==apc_command_t.APC_COMMAND_FILETRANSFER) {
				return sendfile(Encoding.UTF8.GetString(arg), mode);
			} else if(com==apc_command_t.APC_COMMAND_ERROR) {
				putint(ref outblock,(uint)com,0);
				transferblock();
			} else if(com==apc_command_t.APC_COMMAND_ENTERDIR) {
				if( (Encoding.UTF8.GetCharCount(arg)+1+4) <= (TRANSFERBLOCKSIZE/2-4) ) {
					putint(ref outblock,(uint)com,0);
					Buffer.BlockCopy(arg,0,outblock,4,Math.Min(TRANSFERBLOCKSIZE/2-4-4,arg.Length));
					putint(ref outblock,mode,TRANSFERBLOCKSIZE/2-4);
					transferblock();
				} else {
					Console.WriteLine("Too long dirname!");
					return 1;
				}
			} else if(com==apc_command_t.APC_COMMAND_LEAVEDIR) {
				putint(ref outblock,(uint)com,0);
				transferblock();
			} else if(com==apc_command_t.APC_COMMAND_FINISHED) {
				putint(ref outblock,(uint)com,0);
				transferblock();
			} else if(com==apc_command_t.APC_COMMAND_NOP) {
				putint(ref outblock,(uint)com,0);
				transferblock();
			} else if(com==apc_command_t.APC_COMMAND_ABORT) {
				putint(ref outblock,(uint)com,0);
				transferblock();
			} else if(com==apc_command_t.APC_COMMAND_SKIPFILE) {
				putint(ref outblock,(uint)com,0);
				transferblock();
			} else {
				Console.WriteLine( "Unknown command "+com+" to send!");
				return 1;
			}
			return 0;
		}


		
		static int receive(){
			byte[] arg = null;
			String argStr = null, tstr = null;
			apc_command_t com = apc_command_t.APC_COMMAND_FILETRANSFER;
			int ret=0;
			double msecs;
			double d1;
			uint mode = 0;
			tfl.files = 0;
			tfl.bytes = 0;
			
			do {
				if(recvcommand(ref com,ref arg,ref mode)==0) {
					if(com==apc_command_t.APC_COMMAND_ENTERDIR) {
						if(arg!=null) {
							argStr = Encoding.UTF8.GetString(arg);
							argStr=argStr.Substring(0,argStr.IndexOf('\0'));
							basedir = basedir+argStr+"\\";
								if (!Directory.Exists(basedir)){
									try{
										Directory.CreateDirectory(basedir);
									}
									catch(Exception e){
										Console.WriteLine("Can't create dir "+basedir+", aborting!\r\n"+e.ToString());
										ret = 1;
										break;
									}
								} else if (File.Exists(basedir)){
									Console.WriteLine("Want to create dir "+basedir+" but there is already a file, aborting!");
									ret = 1;
									break;
								}
						}
					} else if(com==apc_command_t.APC_COMMAND_LEAVEDIR) {
						tstr = basedir.Substring(0,basedir.Length-1);
						if (tstr.LastIndexOf("\\") > 0)
							basedir = tstr.Substring(0,tstr.LastIndexOf("\\")+1);
						else
							basedir = null;
					} else if(com==apc_command_t.APC_COMMAND_ERROR) {
						break;
					}
				} else {
					Console.WriteLine("Transfer aborted");
					break;
				}
			} while(com!=apc_command_t.APC_COMMAND_FINISHED);
			
			if ( ret == 0 ) {
				Console.WriteLine("Whole transfer completed!");
				
				if ( tfl.files > 1 ) {
					msecs = tfl.tvs.TotalMilliseconds;
					d1 = tfl.bytes;
					d1 *= 1000;
					d1/=msecs;
					d1/=1024;
					Console.WriteLine("  Files received: "+tfl.files);
					Console.WriteLine("  Bytes with average rate: "+tfl.files+" @ "+d1+" KB/s");
				}
			} else if ( ret > 1 ) {
				sendcommand( apc_command_t.APC_COMMAND_ABORT, null , 0 );
			}
			return ret;
		}
		
		static int sendRec(string dirstr){
			int ret=0;
			int sended=0;
			apc_command_t com;
			int nm = 65535;
			String filepart = null;
			
			if (Directory.Exists(dirstr)){
				/* Send EnterDir with filepart */
				if (dirstr.EndsWith("\\"))
					dirstr = dirstr.Substring(0,dirstr.Length-2);
				filepart = dirstr.Substring(3);
				if (filepart.IndexOf("\\")> 0){
					filepart = filepart.Substring(filepart.LastIndexOf("\\")+1);
					ret=sendcommand(apc_command_t.APC_COMMAND_ENTERDIR,Encoding.UTF8.GetBytes(filepart+"\0"),(uint)nm);
					com = (apc_command_t)getint(inblock,0);
					if ( com == apc_command_t.APC_COMMAND_ABORT ) ret = -1;
					sended=1;
				} else{
					if (filepart.Length > 0) {
						ret=sendcommand(apc_command_t.APC_COMMAND_ENTERDIR,Encoding.UTF8.GetBytes(filepart+"\0"),(uint)nm);
						com = (apc_command_t)getint(inblock,0);
						if ( com == apc_command_t.APC_COMMAND_ABORT ) ret = -1;
						sended=1;
					}
				}
				
				if (ret == 0){
					foreach (String dir in Directory.GetDirectories(dirstr)){
						ret = sendRec(dir);
						if (ret != 0)
							break;
					}
					foreach (String file in Directory.GetFiles(dirstr)){
						ret = sendRec(file);
						if (ret != 0)
							break;
					}
				}
				
				/* Send LeaveDir */
				if ( ( sended == 1 ) && ( ret == 0 ) ) {
					ret=sendcommand(apc_command_t.APC_COMMAND_LEAVEDIR,null,0);
					com = (apc_command_t) getint(inblock,0);
					if ( com == apc_command_t.APC_COMMAND_ABORT ) ret = -1;
				}
				else {
					Console.WriteLine("can't open dir "+dirstr);
				}
			}
			else if (File.Exists(dirstr)){
				ret=sendcommand(apc_command_t.APC_COMMAND_FILETRANSFER,Encoding.UTF8.GetBytes(dirstr),(uint)nm);
			}
			else {
				Console.WriteLine(" no dir information for "+dirstr);
			}
			return ret;

		}
		
		static int send(string[] args){
			int ret = 0;
			double msecs;
			double d1;
			
			/* init transfer log */
			tfl.files = 0;
			tfl.bytes = 0;
			tfl.tvs = new TimeSpan(0);
			
			foreach (String arg in args){
				if (arg.ToLowerInvariant().StartsWith("port:"))
					continue;
				ret = sendRec(arg);
			}
			
			if (ret == 0){
				sendcommand( apc_command_t.APC_COMMAND_FINISHED, null , 0);
				Console.WriteLine("Whole transfer completed!");
				
				if ( tfl.files > 1 ) {
					msecs = tfl.tvs.TotalMilliseconds;
					d1=tfl.bytes;
					d1*=1000;
					d1 /= msecs;
					d1/=1024;
					Console.WriteLine("  Files sent: "+tfl.files );
					Console.WriteLine(String.Format("  Bytes with average rate: "+tfl.bytes+" @ {0:0.00} KB/s", d1));
				}
			}
			else{
				Console.WriteLine("Transfer aborted!\n");
			}
			return 0;
		}
		static bool IsUserAdministrator()
		{
			//bool value to hold our return value
			bool isAdmin;
			try
			{
				//get the currently logged in user
				WindowsIdentity user = WindowsIdentity.GetCurrent();
				WindowsPrincipal principal = new WindowsPrincipal(user);
				isAdmin = principal.IsInRole(WindowsBuiltInRole.Administrator);
			}
			catch (UnauthorizedAccessException)
			{
				isAdmin = false;
			}
			catch (Exception)
			{
				isAdmin = false;
			}
			return isAdmin;
		}
	}
}